@@ -114,7 +114,7 @@ use parent qw/Plack::Component/;
 
 binmode STDERR, ":utf8"; 
 
-our $VERSION = '0.09';
+our $VERSION = '0.10';
 
 =pod
 
@@ -308,10 +308,13 @@ sub service {
     my %names;
     for my $key (sort keys %$parameters) {
         $names{lc($key)} = $key;
+        $parameters->{$key} = decode utf8 => $parameters->{$key};
     }
 
     my $post = $names{postdata} // $names{'xforms:model'};
-    $post = $post ? $parameters->{$post} : encode($request->content_encoding // 'UTF-8', $request->content);
+    $post = $post ? $parameters->{$post} : $request->content;
+    my $encoding = $request->content_encoding // 'utf8';
+    $post = decode $encoding => $post;
 
     if ($post) {
         my $parser = XML::LibXML->new(no_blanks => 1);
@@ -408,8 +411,8 @@ Stream an error report as an XML message of type
 =cut
 
 sub error {
-    my ($responder, $msg) = @_;
-    my $writer = Geo::OGC::Service::XMLWriter::Caching->new();
+    my ($responder, $msg, $headers) = @_;
+    my $writer = Geo::OGC::Service::XMLWriter::Caching->new($headers);
     $writer->open_element('ExceptionReport', { version => "1.0" });
     my $attributes = { exceptionCode => $msg->{exceptionCode} };
     my $content;
@@ -484,7 +487,11 @@ sub CORS {
             push @cors, ('Access-Control-'.$key => $val);
         }
     } else {
-        @cors = ('Access-Control-Allow-Origin' => $config);
+        $default{'Allow-Origin'} = $config;
+        for my $key (keys %default) {
+            my $val = $default{$key};
+            push @cors, ('Access-Control-'.$key => $val);
+        }
     }
     return @cors;
 }
@@ -524,14 +531,10 @@ elements.
 
 sub Operation {
     my ($self, $writer, $operation, $protocols, $parameters) = @_;
-    my @parameters;
+    my @p;
     for my $p (@$parameters) {
         for my $n (keys %$p) {
-            my @values;
-            for my $v (@{$p->{$n}}) {
-                push @values, ['ows:Value', $v];
-            }
-            push @parameters, ['ows:Parameter', {name=>$n}, \@values];
+            push @p, [$self->Parameter($n, $p->{$n})];
         }
     }
     my $constraint;
@@ -541,7 +544,19 @@ sub Operation {
         if $protocols->{Get};
     push @http, [ 'ows:Post' => { 'xlink:type'=>'simple', 'xlink:href'=>$self->{config}{resource} } ]
         if $protocols->{Post};
-    $writer->element('ows:Operation' => { name => $operation }, [['ows:DCP' =>['ows:HTTP' => \@http ]], @parameters]);
+    $writer->element('ows:Operation' => { name => $operation }, [['ows:DCP' => ['ows:HTTP' => \@http ]], @p]);
+}
+
+sub Parameter {
+    my ($self, $name, $values) = @_;
+    my $wrap = $values->[0] =~ /allowedvalues/i;
+    $values = $values->[1] if $wrap;
+    my @values;
+    for my $value (@$values) {
+        push @values, ['ows:Value' => $value];
+    }
+    @values = ('ows:AllowedValues' => [@values]) if $wrap;
+    return ('ows:Parameter', { name => $name }, \@values);
 }
 
 =pod
@@ -744,6 +759,7 @@ Geo::OGC::Service::XMLWriter::Streaming.
 
 package Geo::OGC::Service::XMLWriter::Caching;
 use Modern::Perl;
+use Encode qw(decode encode is_utf8);
 
 our @ISA = qw(Geo::OGC::Service::XMLWriter);
 
@@ -788,8 +804,8 @@ sub stream {
     $writer->write($self->{declaration});
     my $xml = '';
     for my $line (@{$self->{cache}}) {
-        $writer->write($line);
-        $xml .= $line;
+        $writer->write(encode utf8 => $line);
+        $xml .= $line if $debug;
     }
     $writer->close;
     if ($debug) {
